#include <iostream>
#include <time.h>
#include "brute.h"
#include "heuristic.h"
#include "dynamic.h"

using namespace std;

int **items, id, n, maxWeight;

void printHelp(char *prog) {
		cout << "Pouziti: " << prog  << " ( bruteFast | bruteSlow | brutePrune | heuristicRate | heurisiticTest | dynamic | compareRate | compareTest )";
		cout << " loop filename [print]" << endl;
		cout << "bruteFast - rychle prohledavani hrubou silou (for cyklus)" << endl;
		cout << "bruteSlow - pomalejsi prohledavani hrubou silou s pomoci zasobniku" << endl;
		cout << "brutePrune - prohledavani hrubou silou metodou vetveni a mezi" << endl;
		cout << "heuristicRate - heuristika s pomerem cena/vaha" << endl;
		cout << "heuristicTest - heuristika s pomerem cena/vaha s testem nejcenejsi veci" << endl;
		cout << "dynamic - prohledavani s pomoci dynamickeho programovani" << endl;
		cout << "compareRate - porovnani bruteFast s heuristicRate" << endl;
		cout << "compareTest - porovnani bruteFast s heuristicTest" << endl;
		exit(1);
}

char *readInt(char *str, int *res);
int loadData(FILE * fin);

enum Kind {bf, bs, bp, hc, ht, dyn, cr, ct};

int main(int argc, char* argv[]) {
		if(argc < 4) printHelp(argv[0]);

		Kind kind;
		if(strcmp(argv[1], "bruteFast") == 0) kind = bf;
		else if(strcmp(argv[1], "bruteSlow") == 0) kind = bs;
		else if(strcmp(argv[1], "brutePrune") == 0) kind = bp;
		else if(strcmp(argv[1], "heuristicRate") == 0) kind = hc;
		else if(strcmp(argv[1], "heuristicTest") == 0) kind = ht;
		else if(strcmp(argv[1], "dynamic") == 0) kind = dyn;
		else if(strcmp(argv[1], "compareRate") == 0) kind = cr;
		else if(strcmp(argv[1], "compareTest") == 0) kind = ct;
		else printHelp(argv[0]);

		int count = 1;
		readInt(argv[2], &count);
		bool print = false;
		if((argc > 4) && (strcmp(argv[4], "print") == 0)) {
				print = true;
				count = 1;
		}

		int ins = 0;
		float maxError = 1;
		float avgError = 0;
		FILE *fin;
		if(fopen_s(&fin, argv[3], "r") != 0) {
				cout << "Nelze otevrit " << argv[3] << endl;
				exit(1);
		}

		time_t begin;
		time(&begin);
		for(int loop = count-1; loop >= 0; loop--) {
				ins = 0;

				while(!feof(fin)) {
						if(loadData(fin) == 0) continue;

						switch(kind) {
								case bf:
										bruteFast(print);
										break;
								case bs:
										bruteSlow(print);
										break;
								case bp:
										brutePrune(print);
										break;
								case hc:
										heuristicRate(print);
										break;
								case ht:
										heuristicTest(print);
										break;
								case cr:	{
										int a = dynamic(false);
										int b = heuristicRate(false);
										float p = (float) b / (float) a;
										if(p < maxError) maxError = p;
										avgError += p;
										loop = -1;
										}	break;
								case ct:	{
										int a = dynamic(false);
										int b = heuristicTest(false);
										float p = (float) b / (float) a;
										if(p < maxError) maxError = p;
										avgError += p;
										loop = -1;
										}	break;
								case dyn:
										dynamic(print);
										break;
								default:
										cout << "Chyba v programu" << endl;
						}

						for(int i = n-1; i >= 0; i--) delete [] items[i];
						delete [] items;
						ins++;
				}
				rewind(fin);
		}
		time_t ctime = time(NULL) - begin;

		fclose(fin);

		cout << endl << "Metoda: " << argv[1] << endl;
		cout << "Soubor: " << argv[3] << endl;
		cout << "Pocet instanci: " << ins << endl;
		if((kind == ct) || (kind == cr)) {
				cout << "Maximalni chyba: " << maxError << endl;
				cout << "Prumerna chyba: " << avgError / ins << endl;
		} else {
				cout << "Pocet opakovani: " << count << endl;
				cout << "celkovy cas: " << ctime << "s" << endl;
				cout << endl << "Prumerny cas na jednu instanci: " << (float) ctime / count / ins << "s" << endl;
		}
		return 0;
}

char *readInt(char *str, int *res) {
		*res = 0;

		while((str[0] == 0) || (str[0] == 10) || (str[0] == 13) || (str[0] == 32)) str++;

		while((str[0] != 0) && (str[0] != 10) && (str[0] != 13) && (str[0] != 32)) {
				if((str[0] >= '0') && (str[0] <= '9')) {
						*res = (*res * 10) + (str[0] - '0');
						str++;
				} else {
						cout << "Neplatny znak ve vstupnim souboru:" << str[0] << "|" << endl;
						exit(1);
				}
		}
		return str;
}

int loadData(FILE * fin) {
		char line[1024] = "";
		char *act = line;
		fgets(line, 1024, fin);

		if((strlen(line) > 1024) || (strlen(line) < 5)) {
				return 0;
		}

		act = readInt(act, &id);
		act = readInt(act, &n);
		if(n > (sizeof(unsigned long long) * 8)) {
				cout << "Prilis mnoho polozek" << endl;
				exit(1);
		}
		act = readInt(act, &maxWeight);
		items = new int * [n];
		for(int i = n-1; i >= 0; i--) items[i] = new int[2];

		int i = 0;
		while(i < n) {
				act = readInt(act, &items[i][0]); // vaha
				act = readInt(act, &items[i][1]); // cena
				i++;
		}

		return 1;
}
