#include <iostream>
#include "brute.h"
#include "Stack.h"
#include "StackItem.h"
#include "global.h"

using namespace std;

static int maxPrice;

bool isOptimal(unsigned long long test);
bool isOptimal(int weight, int prize);
int getWeight(unsigned long long item);
int getPrize(unsigned long long item);
bool canBeMaxPrize(unsigned long long item, int last);
bool canBeMaxPrize(int prize, int last);

int bruteFast(bool print) {
		unsigned long long best = 0;
		unsigned long long max = getLongLong(n) - 1;
		maxPrice = 0;

		for(unsigned long long i = max; i != -1; i--) {
				if(isOptimal(i)) {
						best = i;
				}
		}

		if(print) printSolution(best, maxPrice, max+1);

		return maxPrice;
}

int bruteSlow(bool print) {
		Stack stack;

		maxPrice = 0;
		unsigned long long best = 0;

		stack.push(best, 0);

		while(!stack.empty()) {
				StackItem *item = stack.pop();

				unsigned long long test = item->getConfig();
				if(isOptimal(test)) {
						best = test;
				}
				int first = item->getLastChange();
				for(int i = first; i < n; i++) {
						stack.push(item->set(i), i+1);
				}

				delete item;
		}

		if(print) printSolution(best, maxPrice, getLongLong(n));

		return maxPrice;
}

int brutePrune(bool print) {
		Stack stack;

		maxPrice = 0;
		unsigned long long best = 0;
		unsigned long long step = 0;

		stack.push(best, 0);

		while(!stack.empty()) {
				step++;
				StackItem *item = stack.pop();

				unsigned long long test = item->getConfig();
				int weight = getWeight(test);
				int prize = getPrize(test);
				if(isOptimal(weight, prize)) {
						best = test;
				}
				int first = item->getLastChange();
				if((weight < maxWeight) && canBeMaxPrize(test, first)) {
						for(int i = first; i < n; i++) {
								stack.push(item->set(i), i+1);
						}
				}

				delete item;
		}


		if(print) printSolution(best, maxPrice, step);

		return maxPrice;
}

int getWeight(unsigned long long item) {
		int weight = 0;

		for(int j = n-1; j >= 0; j--) {
				if(item & getLongLong(j)) {
						weight += items[j][0];
				}
		}
		return weight;
}

int getPrize(unsigned long long item) {
		int prize = 0;

		for(int j = n-1; j >= 0; j--) {
				if(item & getLongLong(j)) {
						prize += items[j][1];
				}
		}
		return prize;
}

bool isOptimal(unsigned long long test) {
		int weight = 0;
		int prize = 0;

		for(int j = n-1; j >= 0; j--) {
				if(test & getLongLong(j)) {
						weight += items[j][0];
						prize += items[j][1];
				}
		}
		if((weight <= maxWeight) && (prize > maxPrice)) {
				maxPrice = prize;
				return true;
		}
		return false;
}

bool isOptimal(int weight, int prize) {
		if((weight <= maxWeight) && (prize > maxPrice)) {
				maxPrice = prize;
				return true;
		}
		return false;
}


bool canBeMaxPrize(unsigned long long item, int last) {
		int prize = 0;

		for(int j = last-1; j >= 0; j--) {
				if(item & getLongLong(j)) {
						prize += items[j][1];
				}
		}
		for(int j = last; j < n; j++) {
				prize += items[j][1];
		}
		return prize > maxPrice;
}

bool canBeMaxPrize(int prize, int last) {
		for(int j = last; j < n; j++) {
				prize += items[j][1];
		}
		return prize > maxPrice;
}